import json
import logging
from datetime import datetime, timezone
from functools import cached_property
from typing import Any, AsyncIterator, Coroutine, Dict, Iterator, List, Optional, Union, cast
from uuid import uuid4

import httpx
from openai import APIStatusError, AsyncStream
from openai import AsyncOpenAI as BaseAsyncOpenAI
from openai import OpenAI as BaseOpenAI
from openai._streaming import Stream, extract_stream_chunk_type
from openai._types import ResponseT
from openai.resources.chat import AsyncChat, AsyncCompletions, Chat, Completions
from openai.types.chat import ChatCompletion, ChatCompletionChunk, ChatCompletionMessageParam
from openai.types.chat.chat_completion_chunk import Choice
from openai.types.completion_usage import CompletionUsage
from pydantic import BaseModel, ConfigDict, Field

from baserun.api import ApiClient
from baserun.integrations.integration import Integration
from baserun.mixins import ClientMixin, CompletionMixin
from baserun.models.evals import CompletionEval, TraceEval
from baserun.models.tags import Tag
from baserun.utils import copy_type_hints

logger = logging.getLogger(__name__)


class WrappedChatCompletion(ChatCompletion, CompletionMixin):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    client: "WrappedOpenAIBaseClient"
    name: str
    error: Optional[str] = None
    trace_id: str
    completion_id: str
    template: Optional[str]
    start_timestamp: datetime
    first_token_timestamp: datetime
    end_timestamp: datetime
    input_messages: List[ChatCompletionMessageParam] = Field(default_factory=list)
    config_params: Dict[str, Any] = Field(default_factory=dict)
    tool_results: List[Any] = Field(default_factory=list)
    tags: List[Tag] = Field(default_factory=list)
    evals: List[CompletionEval] = Field(default_factory=list)

    def submit_to_baserun(self):
        self.client.api_client.submit_completion(self)


class WrappedStreamBase(CompletionMixin, BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    client: "WrappedOpenAIBaseClient"
    id: Optional[str] = None
    name: str
    error: Optional[str] = None
    trace_id: str
    completion_id: str
    template: Optional[str] = None
    start_timestamp: datetime
    end_timestamp: Optional[datetime] = None
    first_token_timestamp: Optional[datetime] = None
    input_messages: List[ChatCompletionMessageParam] = Field(default_factory=list)
    config_params: Dict[str, Any] = Field(default_factory=dict)
    tool_results: List[Any] = Field(default_factory=list)
    tags: List[Tag] = Field(default_factory=list)
    evals: List[CompletionEval] = Field(default_factory=list)

    # Used for evals on streaming responses
    captured_choices: List[Choice] = Field(default_factory=list)

    def __init__(self, *args, **kwargs):
        CompletionMixin.__init__(self)
        BaseModel.__init__(self, **kwargs)
        # Get rid of special kwargs that we take but OpenAI doesn't
        kwargs.pop("name", None)
        kwargs.pop("error", None)
        kwargs.pop("completion_id", None)
        kwargs.pop("trace_id", None)
        kwargs.pop("template", None)
        kwargs.pop("config_params", None)
        kwargs.pop("start_timestamp", None)
        kwargs.pop("end_timestamp", None)
        kwargs.pop("first_token_timestamp", None)
        kwargs.pop("tool_results", None)
        kwargs.pop("tags", None)
        kwargs.pop("evals", None)
        kwargs.pop("input_messages", None)
        AsyncStream.__init__(
            self,
            cast_to=extract_stream_chunk_type(AsyncStream[ChatCompletionChunk]),
            *args,
            **kwargs,
        )

    def submit_to_baserun(self):
        if not self.end_timestamp:
            self.end_timestamp = datetime.now(timezone.utc)
        self._client.api_client.submit_stream(self)


class WrappedAsyncStream(WrappedStreamBase, AsyncStream[ChatCompletionChunk]):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._iterator = self.__stream__()

    async def __stream__(self) -> AsyncIterator[ChatCompletionChunk]:
        stream = super().__stream__()
        async for item in stream:
            # TODO? Support n > 1
            self.captured_choices.append(item.choices[0])
            self.id = item.id
            yield item

        self.submit_to_baserun()


class WrappedSyncStream(WrappedStreamBase, Stream[ChatCompletionChunk]):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._iterator = self.__stream__()

    def __iter__(self) -> Iterator[ChatCompletionChunk]:  # type: ignore[override]
        for item in self._iterator:
            yield item

    def __stream__(self) -> Iterator[ChatCompletionChunk]:
        stream = super().__stream__()
        for item in stream:
            # TODO? Support n > 1
            self.captured_choices.append(item.choices[0])
            self.id = item.id
            yield item

        self.submit_to_baserun()


class WrappedCompletions(Completions):
    _client: "WrappedSyncOpenAIClient"

    @copy_type_hints(Completions)
    def create(
        self,
        *args,
        name: Optional[str] = None,
        variables: Optional[Dict[str, str]] = None,
        template: Optional[str] = None,
        **kwargs,
    ) -> Union[WrappedChatCompletion, WrappedSyncStream]:
        messages: List[ChatCompletionMessageParam] = kwargs.pop("messages", [])
        if not name:
            for message in reversed(messages):
                if message.get("content") and message.get("role") == "assistant":
                    name = str(message["content"])[:20]
                    break
        if not name:
            name = "OpenAI Completion"

        try:
            start_timestamp = datetime.now(timezone.utc)
            stream_or_completion = super().create(*args, **kwargs, messages=messages)
            first_token_timestamp = datetime.now(timezone.utc)
            wrapped: Union[WrappedChatCompletion, WrappedSyncStream, None] = None

            if isinstance(stream_or_completion, ChatCompletion):
                wrapped = WrappedChatCompletion(
                    client=self._client,
                    trace_id=self._client.trace_id,
                    name=name,
                    template=template,
                    start_timestamp=start_timestamp,
                    first_token_timestamp=first_token_timestamp,
                    end_timestamp=first_token_timestamp,
                    config_params=kwargs,
                    completion_id=str(uuid4()),
                    input_messages=messages,
                    **stream_or_completion.model_dump(),
                )

            elif isinstance(stream_or_completion, Stream):
                wrapped = WrappedSyncStream(
                    response=stream_or_completion.response,
                    client=self._client,
                    trace_id=self._client.trace_id,
                    name=name,
                    template=template,
                    start_timestamp=start_timestamp,
                    end_timestamp=None,
                    first_token_timestamp=first_token_timestamp,
                    completion_id=str(uuid4()),
                    input_messages=messages,
                    config_params=kwargs,
                )
            else:
                logger.debug(f"Unexpected response type: {type(stream_or_completion)}")
                return stream_or_completion

            if variables:
                for key, value in variables.items():
                    wrapped.variable(key, value)
            else:
                wrapped.submit_to_baserun()

            return wrapped
        except APIStatusError as e:
            self._client.error = json.dumps(
                {
                    "exception": f"{e.__module__}.{e.__class__.__name__}",
                    "code": e.status_code,
                    "body": e.body,
                    "request_id": e.request_id,
                }
            )
            self._client.submit_to_baserun()
            completion_id = str(uuid4())
            wrapped = WrappedChatCompletion(
                id=f"chatcompl-error-{completion_id}",
                client=self._client,
                trace_id=self._client.trace_id,
                name=name,
                template=template,
                start_timestamp=start_timestamp,
                end_timestamp=start_timestamp,
                first_token_timestamp=start_timestamp,
                config_params=kwargs,
                completion_id=completion_id,
                input_messages=messages,
                error=self._client.error,
                object="chat.completion",
                choices=[],
                created=int(datetime.now().timestamp()),
                usage=CompletionUsage(total_tokens=0, completion_tokens=0, prompt_tokens=0),
                **kwargs,
            )
            wrapped.submit_to_baserun()
            raise e


class WrappedAsyncCompletions(AsyncCompletions):
    _client: "WrappedAsyncOpenAIClient"

    @copy_type_hints(AsyncCompletions)
    async def create(
        self,
        *args,
        name: Optional[str] = None,
        variables: Optional[Dict[str, str]] = None,
        template: Optional[str] = None,
        **kwargs,
    ) -> Union[WrappedAsyncStream, WrappedChatCompletion]:
        start_timestamp = datetime.now(timezone.utc)
        stream_or_completion = await super().create(*args, **kwargs)
        first_token_timestamp = datetime.now(timezone.utc)

        messages: List[ChatCompletionMessageParam] = kwargs.pop("messages", [])
        wrapped_completion: Union[WrappedAsyncStream, WrappedChatCompletion, None] = None

        if isinstance(stream_or_completion, ChatCompletion):
            if not name:
                name = (stream_or_completion.choices[0].message.content or "OpenAI Completion")[:20]

            wrapped_completion = WrappedChatCompletion(
                client=self._client,
                trace_id=self._client.trace_id,
                name=name,
                template=template,
                start_timestamp=start_timestamp,
                end_timestamp=first_token_timestamp,
                first_token_timestamp=first_token_timestamp,
                completion_id=str(uuid4()),
                config_params=kwargs,
                input_messages=messages,
                **stream_or_completion.model_dump(),
            )
        else:
            if not name:
                name = "OpenAI Completion"

            wrapped_completion = WrappedAsyncStream(
                response=stream_or_completion.response,
                client=self._client,
                trace_id=self._client.trace_id,
                name=name,
                template=template,
                start_timestamp=start_timestamp,
                end_timestamp=None,
                first_token_timestamp=first_token_timestamp,
                config_params=kwargs,
                input_messages=messages,
                completion_id=str(uuid4()),
            )

        if variables:
            for key, value in variables.items():
                wrapped_completion.variable(key, value)
        else:
            wrapped_completion.submit_to_baserun()

        return wrapped_completion


class WrappedChat(Chat):
    @cached_property
    def completions(self) -> WrappedCompletions:
        return WrappedCompletions(self._client)


class WrappedAsyncChat(AsyncChat):
    @cached_property
    def completions(self) -> WrappedAsyncCompletions:
        return WrappedAsyncCompletions(self._client)


class WrappedOpenAIBaseClient(ClientMixin):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    chat: WrappedChat | WrappedAsyncChat
    integrations: List[Integration]

    def __init__(
        self,
        client: Union[BaseOpenAI, BaseAsyncOpenAI],
        *args,
        name: Optional[str] = None,
        user: Optional[str] = None,
        session: Optional[str] = None,
        trace_id: Optional[str] = None,
        api_client: Optional[ApiClient] = None,
        api_key: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ):
        self.name = name or ".".join([v for v in [client.organization, client.__class__.__name__] if v])
        self.tags: List[Tag] = []
        self.evals: List[TraceEval] = []
        self.client = client
        self.trace_id = trace_id or str(uuid4())
        self._result = None
        self.error: Union[str, None] = None
        self.user = user
        self.session = session
        self.start_timestamp = datetime.now(timezone.utc)
        self.end_timestamp: Union[datetime, None] = None
        self.api_client = api_client or ApiClient(self, api_key=api_key)
        self.metadata = metadata or {}
        self.request_ids: Dict[str, str] = {}
        self.integrations = []

        try:
            from baserun.integrations.llamaindex import LLamaIndexInstrumentation

            self.integrate(LLamaIndexInstrumentation)
        except ImportError:
            pass

    @property
    def result(self):
        return self._result

    @result.setter
    def result(self, value):
        self._result = value
        self.submit_to_baserun()

    def submit_to_baserun(self):
        if self.api_client:
            self.api_client.submit_trace()


class WrappedSyncOpenAIClient(WrappedOpenAIBaseClient, BaseOpenAI):
    chat: WrappedChat

    @copy_type_hints(WrappedOpenAIBaseClient)
    def __init__(
        self,
        client: BaseAsyncOpenAI,
        *args,
        name: Optional[str] = None,
        session: Optional[str] = None,
        user: Optional[str] = None,
        trace_id: Optional[str] = None,
        **kwargs,
    ):
        api_client = kwargs.pop("api_client", None)
        api_key = kwargs.pop("api_key", None)
        WrappedOpenAIBaseClient.__init__(
            self,
            client,
            *args,
            **kwargs,
            name=name,
            user=user,
            session=session,
            trace_id=trace_id,
            api_client=api_client,
            api_key=api_key,
        )
        BaseOpenAI.__init__(self, *args, **kwargs)
        self.chat = WrappedChat(self)

    def _process_response(self, *, response: httpx.Response, cast_to: type[ResponseT], **kwargs) -> ResponseT:
        request_id = response.headers.get("x-request-id")
        processed = super()._process_response(response=response, cast_to=cast_to, **kwargs)
        if response.stream:
            # Streaming responses are handled with _process_response_data
            return processed

        completion_id = response.json().get("id")
        self.request_ids[completion_id] = request_id
        return processed

    def _process_response_data(
        self, *, response: httpx.Response, cast_to: type[ResponseT], data: object, **kwargs
    ) -> ResponseT:
        request_id = response.headers.get("x-request-id")
        data_dict: Dict[str, str] = cast(Dict[str, str], data)
        completion_id = data_dict.get("completion_id")
        if completion_id:
            self.request_ids[completion_id] = request_id

        return super()._process_response_data(response=response, data=data, cast_to=cast_to, **kwargs)


class WrappedAsyncOpenAIClient(WrappedOpenAIBaseClient, BaseAsyncOpenAI):
    chat: WrappedAsyncChat

    @copy_type_hints(WrappedOpenAIBaseClient)
    def __init__(
        self,
        client: BaseAsyncOpenAI,
        *args,
        name: Optional[str] = None,
        session: Optional[str] = None,
        user: Optional[str] = None,
        trace_id: Optional[str] = None,
        **kwargs,
    ):
        api_client = kwargs.pop("api_client", None)
        api_key = kwargs.pop("api_key", None)
        WrappedOpenAIBaseClient.__init__(
            self,
            client,
            *args,
            **kwargs,
            name=name,
            user=user,
            session=session,
            trace_id=trace_id,
            api_client=api_client,
            api_key=api_key,
        )
        BaseAsyncOpenAI.__init__(self, *args, **kwargs)
        self.chat = WrappedAsyncChat(self)

    def _process_response(
        self, *, response: httpx.Response, cast_to: type[ResponseT], **kwargs
    ) -> Coroutine[Any, Any, ResponseT]:
        request_id = response.headers.get("x-request-id")
        processed = super()._process_response(response=response, cast_to=cast_to, **kwargs)
        if response.stream:
            # Streaming responses are handled with _process_response_data
            return processed

        completion_id = response.json().get("id")
        self.request_ids[completion_id] = request_id
        return processed

    def _process_response_data(
        self, *, response: httpx.Response, cast_to: type[ResponseT], data: object, **kwargs
    ) -> ResponseT:
        request_id = response.headers.get("x-request-id")
        data_dict: Dict[str, str] = cast(Dict[str, str], data)
        completion_id = data_dict.get("completion_id")
        if completion_id:
            self.request_ids[completion_id] = request_id

        return super()._process_response_data(response=response, data=data, cast_to=cast_to, **kwargs)


class OpenAI(WrappedSyncOpenAIClient):
    def __init__(self, *args, **kwargs):
        client = BaseOpenAI(*args, **kwargs)
        super().__init__(*args, **kwargs, client=client)


class AsyncOpenAI(WrappedAsyncOpenAIClient):
    def __init__(self, *args, **kwargs):
        client = BaseAsyncOpenAI(*args, **kwargs)
        super().__init__(*args, **kwargs, client=client)



from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any, Dict, Optional
from uuid import uuid4

from pydantic import BaseModel, ConfigDict, Field

if TYPE_CHECKING:
    from baserun.mixins import ClientMixin, CompletionMixin


class CompletionEval(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    id: str = Field(default_factory=lambda: str(uuid4()))
    target: "CompletionMixin" = Field(exclude=True)
    target_type: str
    target_id: str
    timestamp: datetime
    name: Optional[str] = None
    score: Optional[float] = None
    metadata: Dict[str, Any] = Field(default_factory=dict)

    def __init__(self, target: "CompletionMixin", **data):
        super().__init__(
            target_type="completion",
            target_id=target.completion_id,
            **data,
            timestamp=datetime.now(timezone.utc),
            target=target,
        )

    def __bool__(self):
        return self.score == 1

    def model_dump(self, *args, **kwargs):
        dumped = super().model_dump()
        dumped["timestamp"] = dumped.pop("timestamp").isoformat()
        return dumped

    def not_includes(self, expected: str, actual: Optional[str] = None) -> bool:
        result = expected not in (actual or self.message_content())
        self.score = 1 if result else 0
        self.target.submit_to_baserun()
        return result

    def includes(self, expected: str, actual: Optional[str] = None) -> bool:
        result = expected in (actual or self.message_content())
        self.score = 1 if result else 0
        self.target.submit_to_baserun()
        return result

    def message_content(self):
        if hasattr(self.target, "captured_choices") and self.target.captured_choices:
            return "".join([d.delta.content for d in self.target.captured_choices if d.delta.content])
        else:
            return self.target.choices[0].message.content


class TraceEval(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    id: str = Field(default_factory=lambda: str(uuid4()))
    target: "ClientMixin" = Field(exclude=True)
    target_type: str
    target_id: str
    timestamp: datetime
    name: Optional[str] = None
    score: Optional[float] = None
    metadata: Dict[str, Any] = Field(default_factory=dict)

    def __init__(self, target: "ClientMixin", **data):
        super().__init__(
            target_type="trace", target_id=target.trace_id, target=target, **data, timestamp=datetime.now(timezone.utc)
        )

    def __bool__(self):
        return self.score == 1

    def model_dump(self, *args, **kwargs):
        dumped = super().model_dump()
        dumped["timestamp"] = dumped.pop("timestamp").isoformat()
        return dumped

    def not_includes(self, expected: str, actual: Optional[str] = None) -> bool:
        if not actual:
            actual = self.target.result or ""

        result = actual not in expected
        self.score = 1 if result else 0
        self.target.submit_to_baserun()
        return result

    def includes(self, expected: str, actual: Optional[str] = None) -> bool:
        result = expected in (actual or self.target.result or "")
        self.score = 1 if result else 0
        self.target.submit_to_baserun()
        return result



import argparse
import asyncio
import inspect
import json
import logging
import os
import sys
import traceback
from time import sleep

import openai
from openai import AsyncOpenAI, NotFoundError, OpenAI
from openai.types import CreateEmbeddingResponse
from openai.types.chat.chat_completion_message import FunctionCall

from baserun import api, init


def openai_chat(prompt="What is the capital of the US?") -> str:
    client = init(OpenAI(), name="openai_chat")
    completion = client.chat.completions.create(
        name="openai_chat completion", model="gpt-4o", messages=[{"role": "user", "content": prompt}]
    )
    return completion.choices[0].message.content


async def openai_chat_async(prompt="What is the capital of the US?") -> str:
    client = init(AsyncOpenAI(), name="openai_chat_async")
    completion = await client.chat.completions.create(
        name="openai_chat_async completion",
        model="gpt-4o",
        messages=[{"role": "user", "content": prompt}],
    )
    return completion.choices[0].message.content


def openai_chat_import(prompt="What is the capital of the US?") -> str:
    from baserun import OpenAI

    client = OpenAI()
    completion = client.chat.completions.create(model="gpt-4o", messages=[{"role": "user", "content": prompt}])
    return completion.choices[0].message.content


def openai_chat_with_log(prompt="What is the capital of the US?") -> str:
    client = init(OpenAI(), name="openai_chat_with_log")
    completion = client.chat.completions.create(
        name="openai_chat_with_log completion",
        model="gpt-4o",
        messages=[{"role": "user", "content": prompt}],
    )
    content = completion.choices[0].message.content
    command = " ".join(sys.argv)
    completion.log("OpenAI Chat Results", metadata={"command": command})
    return content


def openai_chat_unwrapped(prompt="What is the capital of the US?", **kwargs) -> str:
    client = OpenAI()
    completion = client.chat.completions.create(
        model="gpt-4o", messages=[{"role": "user", "content": prompt}], **kwargs
    )
    return completion.choices[0].message.content


async def openai_chat_unwrapped_async_streaming(prompt="What is the capital of the US?", **kwargs) -> str:
    client = AsyncOpenAI()
    completion_generator = await client.chat.completions.create(
        model="gpt-4o",
        messages=[{"role": "user", "content": prompt}],
        **kwargs,
        stream=True,
    )
    content = ""
    async for chunk in completion_generator:
        if new_content := chunk.choices[0].delta.content:
            content += new_content

    return content


def openai_chat_tools(prompt="Say 'hello world'") -> FunctionCall:
    messages = [{"role": "user", "content": prompt}]

    client = init(OpenAI(), name="openai_chat_tools")
    tools = [
        {
            "type": "function",
            "function": {
                "name": "say",
                "description": "Convert some text to speech",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "text": {"type": "string", "description": "The text to speak"},
                    },
                    "required": ["text"],
                },
            },
        }
    ]
    completion = client.chat.completions.create(
        name="openai_chat_tools completion",
        model="gpt-4o",
        messages=[{"role": "user", "content": prompt}],
        tools=tools,
        tool_choice={"type": "function", "function": {"name": "say"}},
    )
    tool_calls = completion.choices[0].message.tool_calls
    completion.eval("openai_chat_functions.function_call").includes(
        "say",
        actual=json.dumps([tool_call.model_dump() for tool_call in tool_calls]),
    )
    completion.tool_result(tool_calls[0], "success")

    assistant_message = {
        "role": completion.choices[0].message.role,
        "content": completion.choices[0].message.content,
        "tool_calls": [tool_call.model_dump() for tool_call in tool_calls],
    }
    messages.append(assistant_message)
    messages.append(
        {"role": "tool", "content": "wow", "tool_call_id": assistant_message.get("tool_calls")[0].get("id")}
    )
    client.chat.completions.create(
        model="gpt-4o",
        messages=messages,
        tools=tools,
    )

    return tool_calls


def openai_chat_tools_streaming(prompt="Say 'hello world'") -> FunctionCall:
    from baserun import OpenAI

    client = OpenAI()
    tools = [
        {
            "type": "function",
            "function": {
                "name": "say",
                "description": "Convert some text to speech",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "text": {"type": "string", "description": "The text to speak"},
                    },
                    "required": ["text"],
                },
            },
        }
    ]
    completion = client.chat.completions.create(
        model="gpt-4o",
        messages=[{"role": "user", "content": prompt}],
        stream=True,
        tools=tools,
        tool_choice={"type": "function", "function": {"name": "say"}},
    )
    for item in completion:
        print(item)
    return completion


def openai_chat_streaming(prompt="What is the capital of the US?") -> str:
    client = init(OpenAI(), name="openai_chat_streaming")
    completion_generator = client.chat.completions.create(
        name="openai_chat_streaming completion",
        model="gpt-4o",
        stream=True,
        messages=[{"role": "user", "content": prompt}],
    )
    content = ""
    for chunk in completion_generator:
        if new_content := chunk.choices[0].delta.content:
            content += new_content

    return content


async def openai_chat_async_streaming(prompt="What is the capital of the US?") -> str:
    client = init(AsyncOpenAI(), name="openai_chat_async_streaming")
    completion_generator = await client.chat.completions.create(
        name="openai_chat_async_streaming completion",
        model="gpt-4o",
        stream=True,
        messages=[{"role": "user", "content": prompt}],
    )
    content = ""
    async for chunk in completion_generator:
        if new_content := chunk.choices[0].delta.content:
            content += new_content

    return content


def openai_chat_error(prompt="What is the capital of the US?"):
    client = init(OpenAI(), name="openai_chat_error")

    original_api_type = openai.api_type
    try:
        client.chat.completions.create(
            name="openai_chat_error completion",
            model="asdf",
            messages=[{"role": "user", "content": prompt}],
        )
    except NotFoundError as e:
        client.eval("openai_chat_async_streaming.content").includes(e.message, "does not exist")
        return f"Errored with {e} successfully"
    finally:
        openai.api_type = original_api_type


def openai_chat_response_format(prompt="What is the capital of the US?") -> str:
    client = init(OpenAI())
    completion = client.chat.completions.create(
        model="gpt-4-turbo",
        messages=[
            {"role": "system", "content": "Respond to the following question in JSON"},
            {"role": "user", "content": prompt},
        ],
        response_format={"type": "json_object"},
    )
    content = completion.choices[0].message.content
    completion.eval("openai_chat.content").includes("Washington")
    return content


def openai_chat_seed(prompt="What is the capital of the US?") -> str:
    client = init(OpenAI())
    completion = client.chat.completions.create(
        model="gpt-4-turbo",
        messages=[{"role": "user", "content": prompt}],
        seed=1234,
    )
    content = completion.choices[0].message.content
    completion.eval("openai_chat.content").includes("Washington")
    return content


def openai_embeddings(inpt: str = "What is the capital of the US?") -> CreateEmbeddingResponse:
    client = init(OpenAI())
    res = client.embeddings.create(
        input=inpt,
        model="text-embedding-ada-002",
    )
    return res


TEMPLATES = {
    "Question & Answer": [
        {"role": "system", "content": "Respond to all messages in the form of a limerick"},
        {"role": "user", "content": "{question}"},
    ]
}


def use_template(question="What is the capital of the US?", template_name: str = "Question & Answer"):
    unformatted_messages = TEMPLATES.get(template_name)
    prompt = [
        {"role": message["role"], "content": message["content"].format(question=question)}
        for message in unformatted_messages
    ]

    client = init(OpenAI(), name="use_template")
    completion = client.chat.completions.create(
        model="gpt-4-turbo", messages=prompt, template=template_name, name="use_template completion"
    )
    content = completion.choices[0].message.content
    completion.eval("openai_chat.content").includes("Washington")
    return content


async def use_template_async(question="What is the capital of the US?", template_name: str = "Question & Answer"):
    unformatted_messages = TEMPLATES.get(template_name)
    prompt = [
        {"role": message["role"], "content": message["content"].format(question=question)}
        for message in unformatted_messages
    ]

    client = init(AsyncOpenAI(), name="use_template_async")
    completion = await client.chat.completions.create(
        model="gpt-4-turbo", messages=prompt, template=template_name, name="use_template_async completion"
    )
    content = completion.choices[0].message.content
    completion.eval("openai_chat.content").includes("Washington")
    return content


def use_sessions(
    prompt="What is the capital of the US?", user_identifier="example@test.com", session_identifier="session-123"
) -> str:
    client = init(OpenAI(), user=user_identifier, session=session_identifier, name="use_sessions")
    completion = client.chat.completions.create(
        name="use_sessions completion",
        model="gpt-4-turbo",
        messages=[{"role": "user", "content": prompt}],
    )
    return completion.choices[0].message.content


def create_full_trace(question="What is the capital of the US?") -> str:
    client = init(OpenAI(), name="Full Trace")
    client.log("Answering question for customer", metadata={"customer_tier": "Pro"})

    completion = client.chat.completions.create(
        model="gpt-4-1106-preview",
        temperature=0.3,
        user="epeterson",
        variables={"question": question},
        messages=[{"role": "user", "content": question}],
    )
    content = completion.choices[0].message.content

    completion.feedback(
        name="question_feedback",
        score=0.8,
        metadata={"comment": "This is correct but is too concise"},
    )
    completion.eval("Contains answer").includes("Washington")
    completion.log(name="Extracted content", message=content)
    completion.submit_to_baserun()
    client.result = content
    client.feedback("content", 0.9, metadata={"comment": "This is a great answer"})
    client.eval("Contains answer").includes(expected="Washington")
    return content


def use_input_variables():
    client = init(OpenAI())
    completion = client.chat.completions.create(
        model="gpt-4o",
        messages=[{"role": "user", "content": "What is the capital of the US?"}],
        variables={"country": "the US"},
    )
    return completion.choices[0].message.content


def call_function(functions, function_name: str, parsed_args: argparse.Namespace):
    function_to_call = functions.get(function_name)
    if function_to_call is None:
        function_to_call = {f: globals().get(f) for f in globals()}.get(function_name)

    if inspect.iscoroutinefunction(function_to_call):
        if parsed_args.prompt:
            result = asyncio.run(function_to_call(parsed_args.prompt))
        else:
            result = asyncio.run(function_to_call())
    else:
        if parsed_args.prompt:
            result = function_to_call(parsed_args.prompt)
        else:
            result = function_to_call()

    print(result)
    return result


# Allows you to call any of these functions, e.g. python tests/testing_functions.py openai_chat_functions_streaming
if __name__ == "__main__":
    from dotenv import load_dotenv

    load_dotenv()
    openai.api_key = os.environ.get("OPENAI_API_KEY")

    parser = argparse.ArgumentParser(description="Execute a function with a prompt.")
    parser.add_argument("function_to_call", type=str, help="Name of the function to call")
    parser.add_argument("--prompt", type=str, help="Prompt to pass to the function", default=None)

    parsed_args = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG)

    # Resolve the string function name to the function object
    function_name = parsed_args.function_to_call
    global_variables = {f: globals().get(f) for f in globals()}
    traced_functions = {n: f for n, f in global_variables.items() if callable(f) and f.__name__ == "wrapper"}
    if function_name == "all":
        for name, func in traced_functions.items():
            print(f"===== Calling function {name} =====\n")
            try:
                result = call_function(traced_functions, name, parsed_args)
                print(f"----- {name} result:\n{result}\n-----")
            except Exception:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                traceback.print_exception(exc_type, exc_value, exc_traceback)
    else:
        call_function(traced_functions, function_name, parsed_args)

    sleep(4)
    api.stop_worker()

    sleep(1)



from typing import Optional, TypeVar

from baserun.mixins import ClientMixin

T = TypeVar("T")

try:
    from openai import AsyncOpenAI as BaseAsyncOpenAI

    from baserun.wrappers.openai import AsyncOpenAI as WrappedAsyncOpenAI
    from baserun.wrappers.openai import OpenAI as WrappedOpenAI
    from baserun.wrappers.openai import (
        WrappedAsyncOpenAIClient,
        WrappedOpenAIBaseClient,
        WrappedSyncOpenAIClient,
    )

    OpenAI = WrappedOpenAI
    AsyncOpenAI = WrappedAsyncOpenAI

    def init(client: T, name: Optional[str] = None, api_key: Optional[str] = None, **kwargs) -> WrappedOpenAIBaseClient:
        if isinstance(client, BaseAsyncOpenAI):
            return WrappedAsyncOpenAIClient(**kwargs, name=name, api_key=api_key, client=client)
        return WrappedSyncOpenAIClient(**kwargs, name=name, api_key=api_key, client=client)

except ImportError:
    # TODO: Handle anthropic and other providers here (in case they've installed any of those alongside openai)

    class BaseOpenAI:
        pass

    class WrappedSyncOpenAIClient(ClientMixin):  # type: ignore
        pass

    class WrappedAsyncOpenAIClient(ClientMixin):  # type: ignore
        pass

    def init(client: T, name: Optional[str] = None, api_key: Optional[str] = None, **kwargs) -> WrappedOpenAIBaseClient:
        raise ImportError("No supported libraries are installed.")



import functools
import json

import tiktoken
from openai.types.chat import ChatCompletionMessageParam


def copy_type_hints(superclass):
    def decorator(func):
        original_func = getattr(superclass, func.__name__)
        func.__annotations__ = original_func.__annotations__

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    return decorator


def deep_merge(dicts):
    result = {}
    for dictionary in dicts:
        for key, value in dictionary.items():
            if key in result:
                # Handling both values are dicts
                if isinstance(result[key], dict) and isinstance(value, dict):
                    result[key] = deep_merge([result[key], value])
                # Handling both values are lists
                elif isinstance(result[key], list) and isinstance(value, list):
                    # If both are lists of dictionaries, merge each corresponding dict
                    if all(isinstance(item, dict) for item in result[key]) and all(
                        isinstance(item, dict) for item in value
                    ):
                        combined = []
                        for item1, item2 in zip(result[key], value):
                            if isinstance(item1, dict) and isinstance(item2, dict):
                                combined.append(deep_merge([item1, item2]))
                        # Handle excess in longer list
                        longer_list = result[key] if len(result[key]) > len(value) else value
                        combined.extend(longer_list[len(combined) :])
                        result[key] = combined
                    else:
                        result[key].extend(value)
                # Handling both values are strings
                elif isinstance(result[key], str) and isinstance(value, str):
                    result[key] += value
                # If the new value is not None and not a list, overwrite
                elif value is not None:
                    result[key] = value
            else:
                result[key] = value
    return result


def count_message_tokens(text: str, encoder="cl100k_base"):
    return len(tiktoken.get_encoding(encoder).encode(text))


def count_prompt_tokens(
    messages: list[ChatCompletionMessageParam],
    encoder="cl100k_base",
):
    """FYI this doesn't count "name" keys correctly"""
    return sum([count_message_tokens(json.dumps(list(m.values())), encoder=encoder) for m in messages]) + (
        len(messages) * 3
    )



from abc import ABC
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from baserun.mixins import ClientMixin


class Integration(ABC):
    def __init__(self, client: "ClientMixin") -> None:
        self.client = client

    def instrument(self) -> None: ...

    def uninstrument(self) -> None: ...



import asyncio
import atexit
import json
import logging
import os
from multiprocessing import Process, Queue
from queue import Empty
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import httpx

from baserun.utils import count_prompt_tokens, deep_merge

if TYPE_CHECKING:
    from baserun.wrappers.openai import (
        WrappedAsyncStream,
        WrappedChatCompletion,
        WrappedOpenAIBaseClient,
        WrappedSyncStream,
    )

logger = logging.getLogger(__name__)

exporter_queue: Queue = Queue()
tasks: List[asyncio.Task] = []  # Make tasks a global variable
exporter_process: Union[Process, None] = None


async def post_and_log_response(
    base_url: str, api_key: str, endpoint: str, data: Dict[str, Any], client: httpx.AsyncClient
):
    response = await client.post(
        f"{base_url}/api/public/{endpoint}",
        headers={"Authorization": f"Bearer {api_key}"},
        json=data,
    )
    if response.status_code != 200:
        logger.warning(f"Response from {base_url}/api/public/{endpoint}: {response.status_code}")
        logger.warning(response.json())


async def worker(queue: Queue, base_url: str, api_key: str):
    logger = logging.getLogger(__name__ + ".worker")
    logger.debug(f"Starting worker with base_url: {base_url}")
    try:
        async with httpx.AsyncClient(http2=True) as client:
            while True:
                try:
                    item: Dict = queue.get_nowait()
                except Empty:
                    await asyncio.sleep(1)
                    continue

                if item is None:
                    break

                try:
                    endpoint = item.pop("endpoint")
                    data = item.pop("data")
                    logger.debug(f"Submitting {data} to Baserun at {base_url}/api/public/{endpoint}")

                    task = asyncio.create_task(post_and_log_response(base_url, api_key, endpoint, data, client))
                    task.add_done_callback(lambda t: tasks.remove(t))
                    task.add_done_callback(lambda t: logger.debug(f"Task {t} finished"))
                    tasks.append(task)
                except Exception as e:
                    logger.warning(f"Failed to submit {endpoint} to Baserun: {e}")

    finally:
        logger.debug(f"Waiting for {len(tasks)} tasks to finish")
        await asyncio.gather(*tasks)
        logger.debug("Exiting worker")
        loop = asyncio.get_running_loop()
        loop.stop()


def run_worker(queue: Queue, base_url: str, api_key: str):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    loop.create_task(worker(queue, base_url, api_key))
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass
    finally:
        loop.close()


def start_worker(base_url: str, api_key: str):
    global exporter_process
    if exporter_process is None:
        exporter_process = Process(
            target=run_worker,
            args=(exporter_queue, base_url, api_key),
        )
        exporter_process.daemon = False
        exporter_process.start()


def stop_worker():
    global exporter_process, exporter_queue, tasks

    # Signal the worker to stop
    if exporter_queue is not None and not getattr(exporter_queue, "_closed"):
        exporter_queue.put(None)

    if exporter_process is not None and exporter_process.is_alive():
        loop = asyncio.get_event_loop()
        if tasks:
            loop.run_until_complete(asyncio.gather(*tasks))

        # Terminate the process
        exporter_process.terminate()
        exporter_process.join()

    # Close and join the queue
    if exporter_queue is not None and not getattr(exporter_queue, "_closed"):
        exporter_queue.close()
        exporter_queue.join_thread()


class ApiClient:
    def __init__(
        self,
        client: Union["WrappedOpenAIBaseClient"],
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
    ):
        self.environment = os.getenv("BASERUN_ENVIRONMENT", os.getenv("ENVIRONMENT", "production"))
        self.client = client

        atexit.register(self.exit_handler)

        start_worker(
            base_url or os.getenv("BASERUN_API_URL") or "https://app.baserun.ai",
            api_key or os.getenv("BASERUN_API_KEY") or "",
        )

    def exit_handler(self, *args) -> None:
        stop_worker()

    def submit_completion(self, completion: "WrappedChatCompletion"):
        dict_items = completion.model_dump()
        dict_items.pop("client", None)
        start_timestamp = dict_items.pop("start_timestamp", None)
        end_timestamp = dict_items.pop("end_timestamp", None)
        first_token_timestamp = dict_items.pop("first_token_timestamp", None)
        input_messages = []
        for message in dict_items.pop("input_messages", []):
            tool_calls = message.pop("tool_calls", [])
            input_messages.append({**message, "tool_calls": [*tool_calls]})

        output = {
            **dict_items,
            "name": completion.name,
            "trace_id": completion.trace_id,
            "tool_results": completion.tool_results,
            "evals": [e.model_dump() for e in completion.evals if e.score is not None],
            "tags": [tag.model_dump() for tag in completion.tags],
            "config_params": dict_items.pop("config_params", {}),
            "environment": self.environment,
            "input_messages": input_messages,
            "trace": self._trace_data(),
            "start_timestamp": start_timestamp.isoformat(),
            "end_timestamp": end_timestamp.isoformat() if end_timestamp else None,
            "first_token_timestamp": first_token_timestamp.isoformat() if first_token_timestamp else None,
            "request_id": self.client.request_ids.get(completion.id),
        }
        logger.debug(f"Submitting completion:\n{json.dumps(output, indent=2)}")
        self._post("completions", output)

    def submit_stream(self, stream: Union["WrappedAsyncStream", "WrappedSyncStream"]):
        if not stream.id:
            return

        merged_choice = deep_merge([c.model_dump() for c in stream.captured_choices])
        merged_choice.pop("delta", None)
        merged_delta = deep_merge([c.delta.model_dump() for c in stream.captured_choices])
        merged_choice["message"] = merged_delta

        output = {
            "id": stream.id,
            "name": stream.name,
            "completion_id": stream.completion_id,
            "model": stream.config_params.get("model"),
            "trace_id": stream.trace_id,
            "tags": [tag.model_dump() for tag in stream.tags],
            "tool_results": stream.tool_results,
            "evals": [e.model_dump() for e in stream.evals if e.score is not None],
            "input_messages": stream.input_messages,
            "choices": [merged_choice],
            "usage": {
                "completion_tokens": len(stream.captured_choices) + 1,
                "prompt_tokens": count_prompt_tokens(stream.input_messages),
                "total_tokens": count_prompt_tokens(stream.input_messages) + len(stream.captured_choices) + 1,
            },
            "config_params": stream.config_params or {},
            "start_timestamp": stream.start_timestamp.isoformat(),
            "end_timestamp": stream.end_timestamp.isoformat() if stream.end_timestamp else None,
            "first_token_timestamp": stream.first_token_timestamp.isoformat() if stream.first_token_timestamp else None,
            "request_id": self.client.request_ids.get(stream.id),
            "environment": self.environment,
            "trace": self._trace_data(),
        }
        logger.debug(f"Submitting streamed completion:\n{json.dumps(output, indent=2)}")
        self._post("completions", output)

    def submit_trace(self):
        output = self._trace_data()

        logger.debug(f"Submitting trace:\n{json.dumps(output, indent=2)}")
        self._post("traces", output)

    def _post(self, endpoint: str, data: Dict[str, Any]):
        exporter_queue.put({"endpoint": endpoint, "data": data})

    def _trace_data(self) -> Dict[str, Any]:
        return {
            "id": self.client.trace_id,
            "name": self.client.name,
            "tags": [tag.model_dump() for tag in self.client.tags],
            "evals": [e.model_dump() for e in self.client.evals if e.score is not None],
            "environment": self.environment,
            "result": self.client.result,
            "start_timestamp": self.client.start_timestamp.isoformat(),
            "end_timestamp": self.client.end_timestamp.isoformat() if self.client.end_timestamp else None,
            "error": self.client.error,
            "end_user_identifier": self.client.user,
            "end_user_session_identifier": self.client.session,
            "metadata": self.client.metadata,
        }



import logging

from baserun import api

logger = logging.getLogger(__name__)


def pytest_addoption(parser):
    parser.addoption("--baserun", action="store_true", help="Enable baserun functionality")


def pytest_runtest_teardown(item, nextitem):
    while not api.exporter_queue.empty():
        api.exporter_queue.get_nowait()

    for task in api.tasks:
        task.cancel()
    api.tasks = []


def pytest_sessionstart(session):
    # This will stop start_worker from starting worker threads
    api.exporter_thread = "Dummy"



import json
from datetime import datetime, timezone
from typing import Any, Dict
from uuid import uuid4

from pydantic import BaseModel, Field


class Tag(BaseModel):
    id: str = Field(default_factory=lambda: str(uuid4()))
    target_type: str
    target_id: str
    tag_type: str
    key: str
    value: str
    timestamp: datetime
    metadata: Dict[str, Any] = Field(default_factory=dict)

    def __init__(self, **data):
        super().__init__(**data, timestamp=datetime.now(timezone.utc))

    def model_dump(self, *args, **kwargs):
        dumped = super().model_dump()
        dumped["timestamp"] = dumped.pop("timestamp").isoformat()
        return dumped


class Variable(Tag):
    def __init__(self, **data):
        key = data.pop("name")
        super().__init__(**data, tag_type="variable", key=key)

    @property
    def name(self) -> str:
        return self.key

    @name.setter
    def name(self, value: str):
        self.key = value


class Transform(Tag):
    def __init__(self, **data):
        value = json.dumps(
            {
                "input": data.pop("input"),
                "output": data.pop("output"),
            }
        )
        super().__init__(**data, tag_type="transform", value=value)

    @property
    def input(self) -> str:
        return json.loads(self.value).get("input")

    @input.setter
    def input(self, input_value: Variable):
        self.value = json.dumps({"input": input_value, "output": self.output})

    @property
    def name(self) -> str:
        return self.key

    @name.setter
    def name(self, value: str):
        self.key = value

    @property
    def output(self) -> str:
        return json.loads(self.value).get("output")

    @output.setter
    def output(self, output_value: Variable):
        self.value = json.dumps(
            {
                "input": self.input,
                "output": output_value,
            }
        )


class Log(Tag):
    def __init__(self, **data):
        key = data.pop("name", data.pop("key", str(uuid4())))
        super().__init__(**data, tag_type="log", key=key)

    @property
    def message(self) -> str:
        return self.value

    @message.setter
    def message(self, value: str):
        self.value = value

    @property
    def name(self) -> str:
        return self.key

    @name.setter
    def name(self, value: str):
        self.key = value



import json
from abc import ABC
from typing import Any, Dict, List, Optional, Type, overload

from openai.types.chat import ChatCompletionMessageToolCall

from baserun.integrations.integration import Integration
from baserun.models.evals import CompletionEval, TraceEval
from baserun.models.tags import Log, Tag, Transform, Variable


class CompletionMixin(ABC):
    tags: List[Tag]
    evals: List[CompletionEval]
    completion_id: str
    tool_results: List[Dict[str, Any]]

    def annotate(self, key: str, value: str, metadata: Optional[Dict[str, Any]] = None):
        self.tags.append(
            Tag(
                target_type="completion",
                target_id=self.completion_id,
                tag_type="annotation",
                key=key,
                value=value,
                metadata=metadata or {},
            )
        )
        self.submit_to_baserun()

    def eval(self, name: str, score: Optional[int] = None, metadata: Optional[Dict[str, Any]] = None) -> CompletionEval:
        evaluator = CompletionEval(target=self, name=name, metadata=metadata or {}, score=score)
        self.evals.append(evaluator)
        return evaluator

    def feedback(self, name: str, score: Any, metadata: Optional[Dict[str, Any]] = None):
        value = json.dumps(score)
        self.tags.append(
            Tag(
                target_type="completion",
                target_id=self.completion_id,
                tag_type="feedback",
                key=name,
                value=value,
                metadata=metadata or {},
            )
        )
        self.submit_to_baserun()

    @overload
    def log(self, message: str, name: str):
        """Given message and name"""

    @overload
    def log(self, message: str, name: str, metadata: Dict):
        """Given message, name, and metadata."""
        pass

    @overload
    def log(self, message: str):
        """Given just message"""
        pass

    def log(self, message: str, name: Optional[str] = None, metadata: Optional[Dict] = None):
        # Allow for transposition of name and metadata
        if isinstance(name, dict):
            metadata = name
            name = None
        elif isinstance(metadata, str):
            name = metadata
            metadata = {}

        self.tags.append(
            Log(
                target_type="completion",
                target_id=self.completion_id,
                key=name or "log",
                value=message,
                metadata=metadata or {},
            )
        )
        self.submit_to_baserun()

    def tool_result(self, tool_call: ChatCompletionMessageToolCall, result: Any):
        self.tool_results.append({"tool_call": tool_call.model_dump(), "result": result})
        self.submit_to_baserun()

    def transform(self, *args, **kwargs):
        self.tags.append(Transform(name=args[0], target_type="trace", target_id=self.completion_id, **kwargs))
        self.submit_to_baserun()

    def variable(self, key: str, value: Any, metadata: Optional[Dict[str, Any]] = None):
        self.tags.append(
            Variable(
                name=key,
                value=value if isinstance(value, str) else json.dumps(value),
                target_type="completion",
                target_id=self.completion_id,
                metadata=metadata or {},
            )
        )
        self.submit_to_baserun()

    def submit_to_baserun(self):
        pass


class ClientMixin(ABC):
    tags: List[Tag]
    evals: List["TraceEval"]
    trace_id: str
    result: Optional[str]
    integrations: List[Integration]

    def annotate(self, key: str, value: str, metadata: Optional[Dict[str, Any]] = None):
        self.tags.append(
            Tag(
                target_type="trace",
                target_id=self.trace_id,
                tag_type="annotation",
                key=key,
                value=value,
                metadata=metadata or {},
            )
        )

    def eval(self, name: str, score: Optional[int] = None, metadata: Optional[Dict[str, Any]] = None) -> "TraceEval":
        evaluator = TraceEval(target=self, name=name, metadata=metadata or {}, score=score)
        self.evals.append(evaluator)
        return evaluator

    def feedback(self, name: str, score: Any, metadata: Optional[Dict[str, Any]] = None):
        value = json.dumps(score)
        self.tags.append(
            Tag(
                target_type="trace",
                target_id=self.trace_id,
                tag_type="feedback",
                key=name,
                value=value,
                metadata=metadata or {},
            )
        )
        self.submit_to_baserun()

    @overload
    def log(self, message: str, name: str):
        """Given message and name"""

    @overload
    def log(self, message: str, name: str, metadata: Dict):
        """Given message, name, and metadata."""
        pass

    @overload
    def log(self, message: str):
        """Given just message"""
        pass

    def log(self, message: str, name: Optional[str] = None, metadata: Optional[Dict] = None):
        # Allow for transposition of name and metadata
        if isinstance(name, dict):
            metadata = name
            name = None
        elif isinstance(metadata, str):
            name = metadata
            metadata = {}

        self.tags.append(
            Log(
                target_type="trace",
                target_id=self.trace_id,
                key=name or "log",
                value=message,
                metadata=metadata or {},
            )
        )
        self.submit_to_baserun()

    def integrate(self, integration_class: Type[Integration]):
        integration = integration_class(client=self)
        integration.instrument()
        self.integrations.append(integration)

    def transform(self, *args, **kwargs):
        transform_input = kwargs.pop("input", None)
        transform_output = kwargs.pop("output", None)
        self.tags.append(
            Transform(
                key=args[0],
                target_type="trace",
                target_id=self.trace_id,
                input=transform_input if isinstance(transform_input, str) else json.dumps(transform_input),
                output=transform_output if isinstance(transform_output, str) else json.dumps(transform_output),
                **kwargs,
            )
        )
        self.submit_to_baserun()

    def variable(self, key: str, value: Any, metadata: Optional[Dict[str, Any]] = None):
        self.tags.append(
            Variable(
                name=key,
                value=value if isinstance(value, str) else json.dumps(value),
                target_type="trace",
                target_id=self.trace_id,
                metadata=metadata or {},
            )
        )
        self.submit_to_baserun()

    def submit_to_baserun(self):
        pass


TraceEval.model_rebuild()
CompletionEval.model_rebuild()



import json
from typing import (
    Any,
    Callable,
    Union,
)

import llama_index.core.instrumentation as instrument
from llama_index.core.instrumentation.event_handlers import BaseEventHandler
from llama_index.core.instrumentation.events import BaseEvent
from llama_index.core.instrumentation.events.retrieval import RetrievalEndEvent, RetrievalStartEvent

from baserun.integrations.integration import Integration
from baserun.mixins import ClientMixin


class BaserunLlamaEventHandler(BaseEventHandler):
    log_func: Callable[[str, Union[str]], Any]
    enabled: bool = False

    def enable(self) -> None:
        self.enabled = True

    def disable(self) -> None:
        self.enabled = False

    @classmethod
    def class_name(cls) -> str:
        return "BaserunLlamaEventHandler"

    def handle_retrieval_start(self, event: RetrievalStartEvent) -> None:
        if isinstance(event.str_or_query_bundle, str):
            self.log_func(event.str_or_query_bundle, "Query for nodes retrieval")
        else:
            self.log_func(event.str_or_query_bundle.to_json(), "Query for nodes retrieval")

    def handle_retrieval_end(self, event: RetrievalEndEvent) -> None:
        self.log_func(json.dumps([node.dict() for node in event.nodes]), "Selected nodes")

    def handle(self, event: BaseEvent, **kwargs) -> None:
        if not self.enabled:
            return

        if isinstance(event, RetrievalStartEvent):
            self.handle_retrieval_start(event)
        if isinstance(event, RetrievalEndEvent):
            self.handle_retrieval_end(event)


class LLamaIndexInstrumentation(Integration):
    def __init__(self, client: ClientMixin) -> None:
        super().__init__(client=client)
        self.event_handler = BaserunLlamaEventHandler(log_func=client.log)
        dispatcher = instrument.get_dispatcher()
        dispatcher.add_event_handler(self.event_handler)

    def instrument(self) -> None:
        self.event_handler.enable()

    def uninstrument(self) -> None:
        self.event_handler.disable()
